package MlFront_Thunk

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module BuildSystems.BusyAsyncBuildSource

Parameters

module S : MUTABLE_STORE
module O : ORIGIN_WITH_DEPTH with type k = S.k
module SM : MUTABLE_STORE_MESSAGES with type k = S.k and type v = S.v
module C : BuildConstraints.MONAD_STATE_ASYNC with type state = S.t and type 'a promise = 'a P.t

Signature

include BUILD_CORE with type k = S.k with type i = S.i with type v = S.v with type vc = S.v C.t with type lifted_vc = S.v C.t with type o = O.t with type store = S.t
Sourcetype k = S.k

The type of build task key which, for build systems, is conventionally a filename.

Sourcetype i = S.i

The type of build information that persists from one invocation of the build system to the next.

Sourcetype v = S.v

The type of build task value which, for build systems, is conventionally the contents of a file.

Sourcetype vc = S.v C.t

The type of applicative/monad-constrained build task value.

Sourcetype lifted_vc = S.v C.t

The type of applicative/monad-constrained build task value after lifting.

Sourcetype o = O.t

The type of answers to the origins of build tasks.

Sourcetype store = S.t

The type of build store which, for build systems, is conventionally the file system.

Sourcemodule type BUILD_TASKS = TASKS with type k = k and type vc = vc and type lifted_vc = lifted_vc and type o = o

The type of build task specific to the build system.

Sourcemodule type MUTABLE_BUILD_STORE = MUTABLE_STORE with type i = i and type k = k and type v = v and type t = store

The type of mutable build store specific to the asynchronous build system.

Sourcetype 'a promise = 'a P.t
Sourceval build_async : (module BUILD_TASKS) -> o -> k -> (module MUTABLE_BUILD_STORE) -> store -> store promise

build tasks key store, given the task description tasks, a target key key, and a store store, returns a new store in which the value of the target key is up to date.