package MlFront_Thunk

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module type Characters.UCHAR_PARSERSource

type 'a t
module Parser : sig ... end
val (let*) : 'a t -> ('a -> 'b t) -> 'b t
val map : ('a -> 'b) -> 'a t -> 'b t
val return : 'a -> 'a t
val fail : Parser.semantic -> 'a t
val (<?>) : 'a t -> string -> 'a t
val no_expectations : 'a t -> 'a t
val located : 'a t -> 'a Fmlib_parse.Located.t t
val get : Parser.state t

The state is read-only (ie. no set function) because the uchar parser is often used in lexers which are restartable (and hence are not supplied a fresh state, so must re-use the old one).

val ucharp : (Uchar.t -> bool) -> string -> Uchar.t t
val uchar : Uchar.t -> Uchar.t t
val string : string -> string t
val (</>) : 'a t -> 'a t -> 'a t
val backtrack : 'a t -> string -> 'a t
val zero_or_more : 'a t -> 'a list t
val one_or_more : 'a t -> ('a * 'a list) t
val skip_zero_or_more : 'a t -> int t
val skip_one_or_more : 'a t -> int t
val one_or_more_separated : ('item -> 'r t) -> ('r -> 'sep -> 'item -> 'r t) -> 'item t -> 'sep t -> 'r t
val expect_end : 'a -> 'a t
val lexer : 'a t -> 'tok -> 'tok t -> (Fmlib_parse.Position.range * 'tok) t