package MlFront_Thunk

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module ThunkLexers.JsonLexerSource

A JSON lexer. It should lex according to the JSON spec ECMA-404, and also accept C-style "// single line comments" and "/* multi-line comments */".

Implementation Notes

Unlike the "simple" JSON example in fmlib, this lexer is Unicode-aware, handles escaped characters, considers horizontal tab 0x09 to be whitespace, accepts decimal digits, has null as a token, and uses separate true from false tokens rather than a Bool token.

Sourcemodule Token : sig ... end
Sourcemodule Token_plus : sig ... end
module CP : sig ... end

Character parser for JSON.

Sourcemodule Lexer : sig ... end