package MlFront_Thunk

  1. Overview
  2. Docs
Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Module ThunkLexers.RangesSource

Sourcetype mapped_uchar_lexer = {
  1. inner_range : Fmlib_parse.Position.range;
    (*

    The range of the raw characters that are or need to be fed to the character_parser. Typically this is the range of raw characters that are inside any surrounding quotes.

    *)
}

Character lexer for quoted or escaped tokens within the restricted domain of characters in the index inner_range.

Sourcetype range_plus =
  1. | Raw_range of Fmlib_parse.Position.range
  2. | Mapped_range of mapped_uchar_lexer
    (*

    A range that is either the conventional "raw" range of a token, where the raw characters that make up the token are equivalent to the characters that make up the token value (that is, the token does not need to be unescaped, unquoted or normalized) -or- that is a range of a token that needs character mapping (unescaping, unquoting, normalization) to convert the raw stream of characters to a token value.

    *)
Sourceval start_of_range_plus : range_plus -> Fmlib_parse.Position.t

A range suitable for displaying.